#include <math.h>

#define ITERATIONS 100
#define TOLERANCE 0.001
#define PI 3.14159265358979323846
#define PIPI 9.8696044010893586188178821328931344231716
#define G 386.40 /* 32.2 converted to inches */
#define EPSILON 0.00192 /* 1.6E-4 converted to inches */
#define V 0.00175248 /* 1.217E-5 converted to inches */

/*
 * Only calculate some values once, for speed.
 */
static double eight_over_pipi_g = 8.0 / PIPI / G;
static double four_over_pi_v = 4.0 / PI / V;
static double twenty1000_epsilon = 20000.0 * EPSILON;
static double one_third = 1.0 / 3.0;
static double one_fifth = 0.20;
static double twelve_cubed = 12.0 * 12.0 * 12.0;


/*
 * Returns the required diameter, or a negative number of the diameter
 * could not be calculated within a reasonable number of iterations.
 */
double diameter(double Q, double L, double h_f)
{
     int iterations = 0;
     double D, newD = 6.0;
     double intermediate1, intermediate2, N_r;

     /*
      * Convert Q, L and h_f to inches.
      */
     Q *= twelve_cubed;
     L *= 12.0;
     h_f *= 12.0;

     /*
      * We could just cram the entire formula for D into one huge
      * calculation, but that's disgusting and gross.  We therefore
      * break it up into several intermediate calculations to make
      * things more clear.
      */
     do {
	  D = newD;
	  /* Q * Q is faster than pow(Q, 2.0) */
	  intermediate1 = eight_over_pipi_g * L * Q * Q / h_f;
	  N_r = four_over_pi_v * Q / D;
	  intermediate2 =  twenty1000_epsilon / D + 10.0E6 / N_r;
	  intermediate2 = pow(intermediate2, one_third) + 1;
	  intermediate1 = intermediate1 * 0.0055 * intermediate2;
	  newD = pow(intermediate1, one_fifth);
	  printf("After %d interation%s, D = %.3f inches.\n", ++iterations,
		 (iterations == 1) ? "" : "s", newD);
     } while ((iterations < ITERATIONS) &&
	      (fabs(newD - D) > TOLERANCE));

     if (fabs(newD - D) > TOLERANCE) {
	  printf("Result did not converge after %d iterations.\n",
		 ITERATIONS);
	  return(-1.0);
     }
     else {
	  printf("Result of %.3f inches after %d iteration%s.\n", newD,
		 iterations, (iterations == 1) ? "" : "s");
	  return(newD * 12.0);
     }

     /*NOTREACHED*/
}

		 
