#include <matrix.h>
#include <stdio.h>

void read_data(int *floors, Matrix *m, Matrix *P, Matrix *omega, Matrix *Phi)
{
     int i, j;
     double tmp;

     if (scanf(" %d", floors) != 1) {
	  fprintf(stderr, "Error reading number of floors.\n");
	  exit(1);
     }

     *m = matrix_create(*floors, *floors);

     for (i = 1; i <= *floors; i++) {
	  for (j = 1; j <= *floors; j++) {
	       if (scanf(" %lf", &tmp) != 1) {
		    fprintf(stderr, "Error reading mass matrix.\n");
		    exit(1);
	       }
	       matrix_set(*m, i, j, tmp);
	  }
     }

     /* stiffness matrix is discarded */

     for (i = 1; i <= *floors; i++) {
	  for (j = 1; j <= *floors; j++) {
	       if (scanf(" %lf", &tmp) != 1) {
		    fprintf(stderr, "Error reading stiffness matrix.\n");
		    exit(1);
	       }
	  }
     }
     
     *P = matrix_create(*floors, 1);

     for (i = 1; i <= *floors; i++) {
	  if (scanf(" %lf", &tmp) != 1) {
	       fprintf(stderr, "Error reading force vector.\n");
	       exit(1);
	  }
	  matrix_set(*P, i, 1, tmp);
     }

     *omega = matrix_create(*floors, 1);

     for (i = 1; i <= *floors; i++) {
	  if (scanf(" %lf", &tmp) != 1) {
	       fprintf(stderr, "Error reading omega squared vector.\n");
	       exit(1);
	  }
	  matrix_set(*omega, i, 1, tmp);
     }
     
     *Phi = matrix_create(*floors, *floors);

     for (i = 1; i <= *floors; i++) {
	  for (j = 1; j <= *floors; j++) {
	       if (scanf(" %lf", &tmp) != 1) {
		    fprintf(stderr, "Error reading Phi matrix.\n");
		    exit(1);
	       }
	       matrix_set(*Phi, i, j, tmp);
	  }
     }

     return;
}


void step_2(Matrix *M, Matrix *K, Matrix *Phi_T, Matrix *omega_T, Matrix Phi,
	    Matrix m, Matrix omega)
{
     matrix_transpose(Phi, Phi_T);
     matrix_transpose(omega, omega_T);

     matrix_multiply(*Phi_T, m, M);
     matrix_multiply(*M, Phi, M);

     matrix_multiply(*omega_T, *M, K);

     return;
}

void step_3(Matrix *F, Matrix Phi_T, Matrix P)
{
     matrix_multiply(Phi_T, P, F);

     return;
}

void step_4(Matrix *eta, Matrix F, Matrix K, Matrix omega, double Omega)
{
     int i, max;
     double Omega_2 = Omega * Omega;

     max = matrix_rows(omega);

     *eta = matrix_create(max, 1);

     for (i = 1; i <= max; i++) {
	  double f = matrix_get(F, i, 1);
	  double k = matrix_get(K, 1, i);
	  double w = matrix_get(omega, i, 1);

	  matrix_set(*eta, i, 1, f / (k * (1.0 - Omega_2 / w)));
     }
}

void results(Matrix F, Matrix K, Matrix omega, Matrix Phi)
{
     Matrix eta = 0;
     int i, max = matrix_rows(Phi);
     double total;
     double Omega;
     
     printf("%-12s", "Excitation");
     for (i = 1; i <= max; i++)
	  printf("N-hat = %-4d", i);
     putchar('\n');

     while (scanf(" %lf", &Omega) == 1) {
	  step_4(&eta, F, K, omega, Omega);

#ifdef DEBUG
	  fprintf(stderr, "After step_4 with Omega = %f, eta =\n", Omega);
	  matrix_print(eta, " %8.2f ", stderr);
	  fputs("\n", stderr);
#endif
	  
	  printf("%-12.3f", Omega);
     
	  total = 0;

	  for (i = 1; i <= max; i++) {
	       total += matrix_get(Phi, max, i) * matrix_get(eta, i, 1);
	       printf("%-12.3e", total);
	  }

	  putchar('\n');

	  matrix_destroy(&eta);
     }
     
     return;
}

main()
{
     int floors;
     Matrix m, P, omega, Phi;
     Matrix M, K, Phi_T, omega_T;
     Matrix F;
     
     m = P = omega = Phi = M = K = Phi_T = omega_T = F = 0;
     
     read_data(&floors, &m, &P, &omega, &Phi);

#ifdef DEBUG
     fputs("After read_data, m =\n", stderr);
     matrix_print(m, " %8.2f ", stderr);
     fputs("P =\n", stderr);
     matrix_print(P, " %8.2f ", stderr);
     fputs("omega =\n", stderr);
     matrix_print(omega, " %8.2f ", stderr);
     fputs("Phi =\n", stderr);
     matrix_print(Phi, " %8.2f ", stderr);
     fputs("\n", stderr);
#endif

     step_2(&M, &K, &Phi_T, &omega_T, Phi, m, omega);

#ifdef DEBUG
     fputs("After step_2, M =\n", stderr);
     matrix_print(M, " %8.2f ", stderr);
     fputs("K =\n", stderr);
     matrix_print(K, " %8.2f ", stderr);
     fputs("Phi_T =\n", stderr);
     matrix_print(Phi_T, " %8.2f ", stderr);
     fputs("omega_T =\n", stderr);
     matrix_print(omega_T, " %8.2f ", stderr);
     fputs("\n", stderr);
#endif

     step_3(&F, Phi_T, P);

#ifdef DEBUG
     fputs("After step_3, F =\n", stderr);
     matrix_print(F, " %8.2f ", stderr);
     fputs("\n", stderr);
#endif

     results(F, K, omega, Phi);

     exit(0);
}
