#include <X11/Xlib.h>
#include <sys/param.h>
/*
 * NOTE:
 *
 * Using the "fast" matrix operations in <matrixP.h> speeds up the
 * program considerably, but the programmer must remember not to use
 * any expressiond with side effects when calling those operations,
 * since they are macros which will cause repeated evaluation of
 * expressions.
 */
#include <matrixP.h>
#include <strings.h>
#include <threedtransform.h>
#include <stdio.h>
#include <stdlib.h>
#include "pset9.h"
#include "service.h"

#define SCALE 4.0
#define DEGREE_INCREMENT 5.0

extern void draw_buttons(winfoptr_t myinfop);



/*
 * points 	The array of points to transform.
 * ret		The return array of transformed points.  It should
 * 		point to 0 or a Matrix of the proper size.
 * ax,ay,az	The angles to rotate by.
 * cx,cy,cz	The centroids about which to rotate.
 * sx,sy,sz	The scales to perform after rotation.
 * tx,ty,tz	The translations to do after scaling.
 */
void transform(Matrix points,
	       Matrix *ret,
	       double ax, double ay, double az,
	       double cx, double cy, double cz,
	       double sx, double sy, double sz,
	       double tx, double ty, double tz)
{
     /*
      * The problem set says that we should produce the total
      * transformation matrix M and then multiple each point by that
      * matrix.  However, since my points are all in one array, I only
      * have to do one transformation, so there is no need to
      * calculate M in advance here.
      */

     threed_translate(points, ret, -cx, -cy, -cz);
     threed_degree_rotate_x(*ret, ret, ax);
     threed_degree_rotate_y(*ret, ret, ay);
     threed_degree_rotate_z(*ret, ret, az);
     threed_scale(*ret, ret, sx, sy, sz);
     threed_translate(*ret, ret, tx, ty, tz);

     return;
}

void find_centroids(Matrix points, double *cx, double *cy, double *cz)
{
     int i, max = matrix_rows(points);
     double xmin, ymin, zmin;
     double xmax, ymax, zmax;

     xmin = xmax = fast_matrix_get(points, 1, 1);
     ymin = ymax = fast_matrix_get(points, 1, 2);
     zmin = zmax = fast_matrix_get(points, 1, 3);

     for (i = 2; i <= max; i++) {
	  double x = fast_matrix_get(points, i, 1);
	  double y = fast_matrix_get(points, i, 2);
	  double z = fast_matrix_get(points, i, 3);
	  xmin = MIN(x,xmin);
	  xmax = MAX(x,xmax);
	  ymin = MIN(y,ymin);
	  ymax = MAX(y,ymax);
	  zmin = MIN(z,zmin);
	  zmax = MAX(z,zmax);
     }

     *cx = (xmax + xmin) / 2.0;
     *cy = (ymax + ymin) / 2.0;
     *cz = (zmax + zmin) / 2.0;

     return;
}



/*
 * Read in a filename after printing a prompt, then open the file with
 * the specified modes.  Exit if EOF or if there is an error reading
 * or if the user hits return or if there is an error opening.
 */
static FILE *get_file(char *message, char *modes)
{
     FILE *file;
     char buffer[MAXPATHLEN];
     char *ptr;
     
     printf("Enter the name of the %s (or enter to quit): ", message);
     if (! fgets(buffer, sizeof(buffer), stdin)) {
	  if (feof(stdin)) {
	       exit(0);
	  }
	  else {
	       fprintf(stderr, "Reading %s ", message);
	       perror("name");
	       exit(1);
	  }
     }

     if (ptr = index(buffer, '\n')) {
	  *ptr = '\0';
     }

     if (*buffer == '\0') {
	  exit(0);
     }

     file = fopen(buffer, modes);
     if (! file) {
	  fprintf(stderr, "Opening ");
	  perror(buffer);
	  exit(1);
     }

     return(file);
}



/*
 * Read the points and joints from a file.
 */
static void read_joints(FILE *file, Matrix *_points, joint **_joints)
{
     int count, retval, i;
     char name[5];
     double x, y, z;
     Matrix points;
     joint *joints;
     
     if ((retval = fscanf(file, " %d", &count)) == -1) {
	  if (feof(file)) {
	       fprintf(stderr, "EOF reading count from joint file.\n");
	  }
	  else {
	       perror("Reading joint file");
	  }
	  exit(1);
     }
     else if (retval == 0) {
	  fprintf(stderr, "Malformed joint file detected reading count.\n");
	  exit(1);
     }

     points = matrix_create(count, MSIZE);
     joints = CALLOC(count+1,joint);
     if (! joints) {
	  fprintf(stderr, "Out of memory allocating for joints.\n");
	  exit(1);
     }

     for (i = 0; i < count; i++) {
	  if ((retval = fscanf(file, " %5s %F %F %F",name, &x, &y, &z))== -1) {
	       if (feof(file)) {
		    fprintf(stderr, "EOF reading joint file at %d.\n", i);
	       }
	       else {
		    perror("Reading joint file");
	       }
	       exit(1);
	  }
	  else if (retval == 0) {
	       fprintf(stderr, "Malformed joint file detected at %d.\n", i);
	       exit(1);
	  }

	  fast_matrix_set(points, i+1, X, x);
	  fast_matrix_set(points, i+1, Y, y);
	  fast_matrix_set(points, i+1, Z, z);
	  fast_matrix_set(points, i+1, H, 1.0);
	  strcpy(joints[i].j_name, name);
	  joints[i].point = i+1;
     }

     joints[i].j_name[0] = '\0';
     
     *_points = points;
     *_joints = joints;

     return;
}



/*
 * Find a joint in the joint list and return the address of its structure.
 */
static joint *find_joint(joint *joints, char *name)
{
     joint *ptr = joints;
     
     while (ptr->j_name[0] && strcmp(ptr->j_name, name))
	  ptr++;

     if (ptr->j_name[0])
	  return(ptr);

     fprintf(stderr, "Unknown joint %s in find_joint.\n", name);
     exit(1);
}



/*
 * Read the members from a file.
 */
static void read_members(FILE *file, joint *joints, member **_members)
{
     int count, i, retval;
     member *members;
     char name[5], joint1[5], joint2[5];
     
     if ((retval = fscanf(file, " %d", &count)) == -1) {
	  if (feof(file)) {
	       fprintf(stderr, "EOF reading count from member file.\n");
	  }
	  else {
	       perror("Reading member file");
	  }
	  exit(1);
     }
     else if (retval == 0) {
	  fprintf(stderr, "Malformed member file detected reading count.\n");
	  exit(1);
     }

     members = CALLOC(count+1, member);
     if (! members) {
	  fprintf(stderr, "Out of memory allocating for joints.\n");
	  exit(1);
     }

     for (i = 0; i < count; i++) {
	  if ((retval = fscanf(file, " %5s %5s %5s", name, joint1, joint2))
	      == -1) {
	       if (feof(file)) {
		    fprintf(stderr, "EOF reading member file at %d.\n", i);
	       }
	       else {
		    perror("Reading member file");
	       }
	       exit(1);
	  }
	  else if (retval == 0) {
	       fprintf(stderr, "Malformed joint file detected at %d.\n", i);
	       exit(1);
	  }

	  strcpy(members[i].m_name, name);
	  members[i].start = find_joint(joints, joint1);
	  members[i].end = find_joint(joints, joint2);
     }

     members[i].m_name[0] = '\0';
     
     *_members = members;

     return;
}
	  
	       

/*
 * All parameters are return parameters
 */
void read_data(Matrix *points, joint **joints, member **members)
{
     FILE *joint_file, *member_file;

     joint_file = get_file("joint data file", "r");
     member_file = get_file("member data file", "r");

     read_joints(joint_file, points, joints);

     read_members(member_file, *joints, members);

     fclose(joint_file);
     fclose(member_file);

     return;
}

/*
 * Takes a button event and pointers to integers containing the
 * current x, y, and z rotation angles.  Modifies them appropriately,
 * or exits if the quit button has been depressed.
 *
 * Note that button presses on the lines surrounding or between
 * buttons are ignored.
 */
void process_event(XButtonEvent event, double *x, double *y, double *z)
{
     int ex = event.x;
     int ey = event.y;
     static int bounds[] = {
	  XOFFSET,
	  XOFFSET + BUTTON_WIDTH,
	  XOFFSET + 2 * BUTTON_WIDTH,
	  XOFFSET + 3 * BUTTON_WIDTH,
	  XOFFSET + 4 * BUTTON_WIDTH,
	  XOFFSET + 5 * BUTTON_WIDTH,
	  XOFFSET + 6 * BUTTON_WIDTH,
	  XOFFSET + 7 * BUTTON_WIDTH
     };
	  
     if (ey <= YOFFSET || ey >= YOFFSET + BUTTON_HEIGHT)
	  return;

     if (ex < bounds[0])
	  return;
     else if (ex > bounds[0] && ex < bounds[1]) {
	  /* quit button */
	  exit(0);
     }
     else if (ex > bounds[1] && ex < bounds[2]) {
	  /* X+ button */
	  *x += DEGREE_INCREMENT;
	  return;
     }
     else if (ex > bounds[2] && ex < bounds[3]) {
	  /* X- button */
	  *x -= DEGREE_INCREMENT;
	  return;
     }
     else if (ex > bounds[3] && ex < bounds[4]) {
	  /* Y+ button */
	  *y += DEGREE_INCREMENT;
	  return;
     }
     else if (ex > bounds[4] && ex < bounds[5]) {
	  /* Y- button */
	  *y -= DEGREE_INCREMENT;
	  return;
     }
     else if (ex > bounds[5] && ex < bounds[6]) {
	  /* Z+ button */
	  *z += DEGREE_INCREMENT;
	  return;
     }
     else if (ex > bounds[6] && ex < bounds[7]) {
	  /* Z- button */
	  *z -= DEGREE_INCREMENT;
	  return;
     }

     return;
}



static void draw_structure(struct info winfo, Matrix points, member *members,
			   double cx, double cy, double cz,
			   double ax, double ay, double az,
			   double scale,
			   double cenx, double ceny)
{
     static Matrix transformed = 0;
     member *ptr;
     
     transform(points, &transformed,
	       ax, ay, az,
	       cx, cy, cz,
	       scale, scale, scale,
	       cenx, ceny, 0.0);

     for (ptr = members; ptr->m_name[0] != '\0'; ptr++) {
	  double h1 = fast_matrix_get(transformed, ptr->start->point, 4);
	  double h2 = fast_matrix_get(transformed, ptr->end->point, 4);
	  XDrawLine(winfo.my_display, winfo.my_drawing, winfo.my_context,
		    (int) (fast_matrix_get(transformed,
					   ptr->start->point, 1)/h1),
		    (int) (fast_matrix_get(transformed,
					   ptr->start->point, 2)/h1),
		    (int) (fast_matrix_get(transformed,
					   ptr->end->point, 1)/h2),
		    (int) (fast_matrix_get(transformed,
					   ptr->end->point, 2)/h2));
     }
     return;
}



main()
{
     Matrix points = 0;
     double cx, cy, cz;
     double ax, ay, az;
     joint *joints;
     member *members;
     struct info winfo;
     XButtonEvent event;
     double cenx = (double) XMAX / 2.0;
     double ceny = (double) YOFFSET / 2.0;
     
     read_data(&points, &joints, &members);
     ax = ay = az = 0;
     find_centroids(points, &cx, &cy, &cz);
     
     winfo.my_drawing = start_graphics(XMAX, YMAX, &winfo.my_display,
				       &winfo.my_context);
     draw_buttons(&winfo);

     while (1) {
	  draw_structure(winfo, points, members, cx, cy, cz, ax, ay, az, SCALE,
			 cenx, ceny);
	  update_graphics();
	  do {
	       get_mouse_event(&event);
	  } while (event.type != ButtonPress); /* ignore ButtonRelease */
	  draw_structure(winfo, points, members, cx, cy, cz, ax, ay, az, SCALE,
			 cenx, ceny);
	  process_event(event, &ax, &ay, &az);
     }
     /* NOTREACHED */
     exit(0);
}

	  
