;;; This Lisp code implements a system that performs computations by
;;; matching patterns of rules against expressions and, if they match,
;;; replacing the matched expression as specified by the rule.

;;; List structure represents mathematical expressions in the usual way.
;;; For example, we may have an expression such as: (* (+ x 3) (+ (* x
;;; x) (+ y 1))).  We identify symbols such as the "x" and "y" as
;;; variables (in the mathematical sense); the numbers are, of course,
;;; constants.  We will only match against operators explicitly.  (If
;;; you read the code carefully, you will see that there is a bit of
;;; "dirt" in the abstractions we use, because (variable? '+) will be
;;; true, even though "+" is not a (mathematical) variable!)


;;; -------------------------------------------

;;; The basic idea behind this simplification system is build a set of
;;; rules, consisting of a pattern and an evaluation skeleton.  Given an
;;; expression, we try to match the pattern to the expression, keeping
;;; track of the assignment of pattern variables to segments of the
;;; expression in a dictionary.  If we can perform such a pattern match,
;;; then we use the skeleton part of the rule to indicate the procedures
;;; that should be evaluated to perform the simplification.

(defparameter *verbose* nil
  "Make non-NIL if you want to see a detailed trace")

;;; A simple matcher is given below, built on the appropriate data
;;; abstractions. 

(defun match (pattern expression dictionary)
  (cond ((eq dictionary 'failed) 'failed)
        ((atom pattern)
         (if (atom expression)
           (if (eql pattern expression)
	       dictionary
	     'failed)
	   'failed))
        ((arbitrary-constant? pattern)
         (if (constant? expression)
           (extend-dictionary pattern expression dictionary)
           'failed))
        ((arbitrary-variable? pattern)
         (if (variable? expression)
           (extend-dictionary pattern expression dictionary)
           'failed))
        ((arbitrary-expression? pattern)
         (extend-dictionary pattern expression dictionary))
        ((atom expression) 'failed)
        (t
	 (match (cdr pattern) (cdr expression)
		(match (car pattern) (car expression) dictionary)))))

;;; Once we have a consistent match, we instantiate the skeleton part of
;;; the rule, evaluating when appropriate.

(defun instantiate (skeleton dictionary)
  (when *verbose*
    (format t "~%   Instantiating ~a with ~a" skeleton dictionary))
  (cond ((atom skeleton) skeleton)
        ((skeleton-evaluation? skeleton)
         (evaluate (evaluation-expression skeleton)
                   dictionary))
        (t (cons (instantiate (car skeleton) dictionary)
		 (instantiate (cdr skeleton) dictionary)))))


;;; The control of the rule base matcher consists of a series of tightly
;;; interwoven recursive loops.  The basic idea is that given an
;;; expression, we apply the simplification rules in order until we either
;;; succeed or run out of rules.  Moreover, if the expression to be
;;; simplified is compound, this control pattern is applied first to the
;;; elementary components of the expression, and then to the resulting
;;; simplified whole expression.  The procedure SIMPLIFIER below
;;; returns a procedure, based on a set of rules, that can be applied to
;;; any expression

(defun simplifier (the-rules)
  (labels
    ((simplify-exp (exp)
       (when *verbose*
         (format t "~% Simplifying ~a" exp))
       (try-rules (if (compound? exp)
                    (simplify-parts exp)
                    exp)))
     (simplify-parts (exp)
       (if (null exp)
         '()
         (cons (simplify-exp (car exp))
               (simplify-parts (cdr exp)))))
     (try-rules (exp)
       (when *verbose*
         (format t "~% Trying rules on ~a" exp))
       (labels ((scan (rules)
                  (if (null rules)
                    exp
                    (let ((dictionary (match (pattern (car rules))
                                             exp
                                             (make-empty-dictionary))))
                      (unless (eq dictionary 'failed)
                        (when *verbose* 
                          (format t "~%  Matched rule ~a to ~a"
                                  (car rules) exp)))
                      (if (eq dictionary 'failed)
                        (scan (cdr rules))
                        (simplify-exp (instantiate (skeleton (car rules))
                                                   dictionary)))))))
         (scan the-rules))))
    #'simplify-exp))

;;; Having built our simplification system, we now must build the data
;;; abstractions on which it rests.  The dictionary is represented as a
;;; list of pairs, each pair consisting of the variable name and its
;;; associated value from the expression.  (We choose to implement this as a
;;; single abstraction, not providing a separate layer for the elements.)

;;; Note that assoc is a primitive Lisp procedure which has the behavior
;;; of scanning a list of pairs, testing a key against the first element
;;; of each pair.  If it finds a pair whose first element is eql to the
;;; key, then that pair is returned.  Otherwise it returns a false value (nil).
;;; We make use of the "pun" here that anything that is not false in Lisp
;;; is true.

(defun make-empty-dictionary () '())

(defun extend-dictionary (pat dat dictionary)
  (let ((vname (variable-name pat)))
    (let ((v (assoc vname dictionary)))
      (cond ((not v)
             (cons (list vname dat) dictionary))
            ((eq (cadr v) dat) dictionary)
            (t 'failed)))))

(defun lookup (var dictionary)
  (let ((v (assoc var dictionary)))
    (if (not v)
        var
        (cadr v))))

;;; Predicates for types of expressions are straightforward.

(defun compound? (exp) (pair? exp))
(defun constant? (exp) (numberp exp))
(defun variable? (exp) (atom exp))

(defun pair? (exp) (consp exp))  ; Common Lisp's version of pair? is consp

;;; A rule is represented by a list of two parts, a pattern and a
;;; skeleton.

(defun pattern (rule) (car rule))
(defun skeleton (rule) (cadr rule))

;;; Within each pattern, explicit objects are represented by themselves,
;;; while generic objects are represented by a list, the first element of
;;; which identifies the type of generic object, and the second of which
;;; identifies the variable name of that object within the pattern.  We
;;; distinguish three types of generic objects.

(defun arbitrary-constant? (pat)
  (if (pair? pat) (eq (car pat) '?c) nil))

(defun arbitrary-expression? (pat)
  (if (pair? pat) (eq (car pat) '?) nil))

(defun arbitrary-variable? (pat)
  (if (pair? pat) (eq (car pat) '?v) nil))

(defun variable-name (pat) (cadr pat))

;;; The skeleton is just an expression with one specially-interpreted
;;; component, a skeleton-evaluation of the form (^ e) where e is an
;;; arbitrary Lisp-like expression, but one in which symbols (other
;;; than for the operator of a compound expression) are taken to be
;;; pattern variables, not ordinary Lisp variables.

(defun skeleton-evaluation? (pat)
  (if (pair? pat) (eq (car pat) '^) nil))

(defun evaluation-expression (evaluation) (cadr evaluation))

;;; The following procedure is included for completeness, but you need not
;;; worry about understanding exactly what it does until later in the
;;; term.  For the purposes of this lecture, you can treat it as an
;;; abstraction that evaluates a given form, relative to the assignment of
;;; values to variables indicated in the dictionary.

(defun evaluate (form dictionary)
  (when *verbose* 
    (format t "~%     Eval ~a with ~a" form dictionary))
  (if (atom form)
      (lookup form dictionary)
      (apply (symbol-function
	      (lookup (car form) dictionary))  ;; use Lisp to do the work
             (mapcar #'(lambda (v) (lookup v dictionary))
                     (cdr form)))))

;;; As an example of a simplification system, the following rules will
;;; create a simple algebra simplifier.

(defparameter algebra-rules
  '(
    ( ((? op) (?c e1) (?c e2))               (^ (op e1 e2))                )
;;  ( ((? op) (? e1) (?c e2))                ((^ op) (^ e2) (^ e1))        )
;; Not all operators are commutative
    ( (+ (? e1) (?c e2))                     (+ (^ e2) (^ e1))             )
    ( (* (? e1) (?c e2))                     (* (^ e2) (^ e1))             )
    ( (+ 0 (? e))                            (^ e)                         )
    ( (* 1 (? e))                            (^ e)                         )
    ( (* 0 (? e))                            0                             )
    ( (* (?c e1) (* (?c e2) (? e3)))         (* (^ (* e1 e2)) (^ e3))      )
    ( (* (? e1) (* (?c e2) (? e3)))          (* (^ e2) (* (^ e1) (^ e3)))  )
    ( (* (* (? e1) (? e2)) (? e3))           (* (^ e1) (* (^ e2) (^ e3)))  )
    ( (+ (?c e1) (+ (?c e2) (? e3)))         (+ (^ (+ e1 e2)) (^ e3))      )
    ( (+ (? e1) (+ (?c e2) (? e3)))          (+ (^ e2) (+ (^ e1) (^ e3)))  )
    ( (+ (+ (? e1) (? e2)) (? e3))           (+ (^ e1) (+ (^ e2) (^ e3)))  )
    ( (+ (* (?c c) (? a)) (* (?c d) (? a)))  (* (^ (+ c d)) (^ a))         )
    ( (* (? c) (+ (? d) (? e)))              (+ (* (^ c) (^ d))
                                                (* (^ c) (^ e)))           )
;; Simplify rational expressions to make trig substitutions work more often.
    ( (/ (? e1) (? e2)) (* (/ 1 (^ e2)) (^ e1)) )
    ( (sin 0) 0 )
    ( (sin (* 1/2 pi)) 1 )
    ( (sin pi) 0)
    ( (sin (* 3/2 pi)) -1 )
    ( (sin (* 2 pi)) 0 )
    ( (cos 0) 1 )
    ( (cos (* 1/2 pi)) 0 )
    ( (cos pi) -1 )
    ( (cos (* 3/2 pi)) 0)
    ( (cos (* 2 pi)) 1 )
    ( (tan (? e1)) (/ (sin (^ e1)) (cos (^ e1))) )
    ( (cot (? e1)) (/ 1 (tan (^ e1))) )
    ( (sec (^ e1)) (/ 1 (cos (^ e1))) )
    ( (csc (^ e1)) (/ 1 (sin (^ e1))) )
    ))

(defparameter algsimp (simplifier algebra-rules))

;;; Exactly the same system can be used to build other symbolic
;;; simplifiers.  The following rules create a simple symbolic
;;; differentiator.

(defparameter deriv-rules
  '(
    ( (dd (?c c) (? v))              0                                 )
    ( (dd (?v v) (? v))              1                                 )
    ( (dd (?v u) (? v))              0                                 )
    ( (dd (+ (? x1) (? x2)) (? v))   (+ (dd (^ x1) (^ v))
                                        (dd (^ x2) (^ v)))             )
    ( (dd (* (? x1) (? x2)) (? v))   (+ (* (^ x1) (dd (^ x2) (^ v)))
                                        (* (dd (^ x1) (^ v)) (^ x2)))  )
;;    ( (dd (** (? x) (?c n)) (? v))   (* (* (^ n) (+ (^ x) (^ (- n 1))))
;;                                        (dd (^ x) (^ v)))              )
;; Calculus bug; should be:
    ( (dd (** (? x) (?c n)) (? v))   (* (* (^ n) (** (^ x) (^ (- n 1))))
                                        (dd (^ x) (^ v)))              )
;; I'm going to list the simple and then more complex form of some
;; rules; the simple form isn't really necessary, but is useful when
;; the output of dsimp is not being fed through an algebraic simplifier.
    ( (dd (** e (?v x)) (?v x)) (** e (^ x)) )
    ( (dd (** e (? u)) (?v x))  (* (** e (^ u)) (dd (^ u) (^ x))) )
    ( (dd (sin (?v x)) (?v x)) (cos (^ x)) )
    ( (dd (sin (? u)) (?v x)) (* (cos (^ u)) (dd (^ u) (^ x))) )
    ( (dd (cos (?v x)) (?v x)) (* -1 (sin (^ x))) )
    ( (dd (cos (? u)) (?v x)) (* -1 (* (sin (^ u)) (dd (^ u) (^ x)))) )
    ( (dd (tan (?v x)) (?v x)) (** (sec (^ x)) 2) )
    ( (dd (tan (? u)) (?v x)) (* (** (sec (^ u)) 2) (dd (^ u) (^ x))) )
    ( (dd (cot (?v x)) (?v x)) (* -1 (** (csc (^ x)) 2)) )
    ( (dd (cot (? u)) (?v x)) (* -1 (* (** (csc (^ u)) 2) (dd (^ u) (^ x)))) )
    ( (dd (sec (?v x)) (?v x)) (* (sec (^ x)) (tan (^ x))) )
    ( (dd (sec (? u)) (?v x)) (* (* (sec (^ u)) (tan (^ u)))
				 (dd (^ u) (^ x))) )
    ( (dd (csc (?v x)) (?v x)) (* -1 (* (csc (^ x)) (cot (^ x)))) )
    ( (dd (csc (? u)) (?v x)) (* -1 (* (* (csc (^ u)) (cot (^ u)))
				       (dd (^ u) (^ x)))) )
    ))

(defparameter dsimp (simplifier deriv-rules))

