;;; -*- Mode: Lisp; package: user -*-

(in-package 'user)

(defparameter *debug* t "Controls tracing from constraint propagation")

(defclass domain-variable ()
  ((name :accessor name :initarg :name :initform (make-symbol "DOMAIN-VAR-"))
   (domain :accessor domain :initarg :domain :initform '(nil t))
   (demons :accessor demons :initform nil)))

(defmethod assigned? ((v domain-variable))
  (null (cdr (domain v))))

(defmethod value ((v domain-variable))
  (cond ((assigned? v) (car (domain v)))
        (t (error "Domain variable ~a has no definite value" (name v)))))

(defmethod assign! ((v domain-variable) val)
  (unless (member val (domain v))
    (error "Tried to assign ~a to ~a, but it's not in the domain."
           val (name v)))
  (when *debug* 
    (format t "~%Assigning ~a to ~a" val (name v)))
  (unless (assigned? v)
    (setf (domain v) (list val))
    (notice-assignment v))
  t)

(defmethod disassign! ((v domain-variable) val)
  ;; note: returns t if it actually disassigned, nil otherwise
  (when *debug*
    (format t "~%Disassigning ~a from ~a" val (name v)))
  (when (member val (domain v))         ; no-op if val not in domain
    (setf (domain v) (remove val (domain v)))
    (when *debug*
      (format t "~%   leaving ~a" (domain v)))
    (cond ((null (domain v))
           (when *debug*
             (format t " ***OVERCONSTRAINED***"))
           (error "***Overconstrained variable ~a" (name v)))
          ((assigned? v)
           (when *debug*
             (format t "     thus assigned ~a to ~a by elimination"
                     (value v) (name v)))
           (notice-assignment v)))))

(defmethod notice-assignment ((v domain-variable))
  (dolist (demon (demons v))
    (funcall demon v)))

(defmethod install-demon ((v domain-variable) demon)
  (push demon (demons v)))

(defun add-constraint (predicate &rest variables)
  (let ((demon (make-constraint-demon predicate variables)))
    (dolist (v variables)
      (install-demon v demon))
    (dolist (v variables)
      (notice-assignment v))))

(defun make-constraint-demon (predicate vars)
  ;; The procedure defined here will be called whenever its argument, v,
  ;; is assigned a specific value.  If all but one of the other vars also
  ;; already has a value, then filter its values to leave only those consistent
  ;; with the values assigned to the others.
  ;; The predicate will be applied to the values of the vars, in the given
  ;; order.
  #'(lambda (v)
      (declare (ignore v))
      ;; Note: v, the var that happens to have been assigned, is of no
      ;; particular interest, except as just another assigned variable.
      (let ((unassigned-vars (remove-if #'assigned? vars)))
        (when (= 1 (length unassigned-vars))
          (filter-domain-values (car unassigned-vars) predicate vars)))))

(defmethod filter-domain-values ((v domain-variable) predicate vars)
  ;; All domain variables in vars except v must, at this point, have values.
  ;; We iterate over the possible values in the domain of v and disassign any
  ;; that are inconsistent (under the predicate) with the other known values.
  (let ((the-vals nil)
        (val-base nil))
    ;; Note: there is a slightly kludgey implementation here to prevent us
    ;; having to keep recomputing the list of values of vars to pass to the
    ;; predicate (once per remaining possible value in the domain of v).
    ;; We create a list of all those values and keep a pointer to the pair
    ;; in the list whose car must hold the value-to-test for v.  Then, for
    ;; each value to test, we just plug it in to the appropriate place and
    ;; call the predicate.
    (dolist (var vars)
      (cond ((not (eq v var)) (push (value var) the-vals))
            (t (push nil the-vals)
               (setq val-base the-vals))))
    (setq the-vals (nreverse the-vals))         ; 'cause push got them backward
    (dolist (val (domain v))
      (setf (car val-base) val)
;      (when *debug*
;	    (format t "~%Calling predicate on ~a" the-vals))
      (unless (apply predicate the-vals)
        (disassign! v val)))))

#|
(defparameter v1 
  (make-instance 'domain-variable 
    :name 'v1
    :domain '(1 2 3)))

(defparameter v2
  (make-instance 'domain-variable 
    :name 'v2
    :domain '(2 3 4)))

(add-constraint #'> v1 v2)

(assign! v1 3)
|#

;;; For (require 'constraint)
(provide 'constraint)
