;;; -*- Mode: Lisp; package: user -*-

(in-package 'user)

#|
This is a somewhat inefficient but fairly clear implementation of Identification
Trees, as described in Winston's notes, Chapter 21.
|#

#|
We represent a table of data using the CLOS class data-table, defined here.  Such
a table identifies the column headings and data values in the rows, as well as
a distinguished column label that is used to uniquely identify any row and another
distinguished column label that points out the dependent variable.

We also give a definition of the SUNBURN data table, reproduced from
Chapter 21 of the notes.
|#

(defclass data-table ()
  (;; names of the columns
   (column-names :accessor column-names :initarg :column-names)

   ;; name of the column that contains the unique identifier of a sample
   (name-column :accessor name-column :initarg :name-column)

   ;; name of the column that contains the dependent variable
   (result-column :accessor result-column :initarg :result-column)

   ;; the content of the data-table: a list of rows
   (rows :accessor rows :initarg :rows)

   ;; cached value of the disorder function applied to this data-table
   (cached-disorder :accessor cached-disorder :initform -1.0)))

(defparameter sunburn
  (make-instance 'data-table
    :column-names 
    	'(	Name	Hair	Height	Weight	Lotion	Result)
    :rows '((	Sarah	Blonde	Average	Light	No	Sunburned)
            (	Dana	Blonde	Tall	Average	Yes	None)
            (	Alex	Brown	Short	Average	Yes	None)
            (	Annie	Blonde	Short	Average	No	Sunburned)
            (	Emily	Red	Average	Heavy	No	Sunburned)
            (	Pete	Brown	Tall	Heavy	No	None)
            (	John	Brown	Average	Heavy	No	None)
            (	Katie	Blonde	Short	Light	Yes	None))
    :name-column 'Name
    :result-column 'Result))

#|
An ID-tree is also a class.  Every non-terminal node of the ID-tree is in fact
just another ID-tree.  Terminal nodes will be represented by a data-table of those
individuals classified at that node.  Each (non-terminal) id-tree
has the attribute on which it splits and an association-list of the values of
that attribute and the subtrees they lead to.  Note: We make use of CLOS' ability
to discriminate between a branch that is another id-tree and one that is a
terminal data-table.  See, for example, the definition of disorder, below.
|#

(defclass id-tree ()
  (;; the attribute on which this node of the tree splits
   (attribute :accessor attribute :initarg :attribute)

   ;; list of attributes not yet used higher up in the tree
   (remaining :accessor remaining :initarg :remaining)

   ;; association list associating values of the attribute with subtrees or leaf
   ;; data-tables.
   (branches :accessor branches :initarg :branches) ; really an alist val/subtree
   ))

(defmethod tell-about-recursive ((tree id-tree) indent)
  (format t "~%~v@t~s split on ~a:" indent tree (attribute tree))
  (let ((new-indent (+ indent 2)))
    (dolist (assn (branches tree))
      (format t "~%~v@t~a:" indent (car assn))
      (tell-about-recursive (cdr assn) new-indent))))

(defmethod tell-about ((tree id-tree))
  (let ((indent 0))
    (format t "~%~v@t~s split on ~a:" indent tree (attribute tree))
    (let ((new-indent (+ indent 2)))
      (dolist (assn (branches tree))
	(format t "~%~v@t~a:" indent (car assn))
	(tell-about-recursive (cdr assn) new-indent)))))

(defmethod tell-about ((d data-table))
  (let ((indent 0))
    (dolist (r (rows d))
      (format t "~%~v@t~a" indent r))))

(defmethod tell-about-recursive ((d data-table) indent)
  (dolist (r (rows d))
    (format t "~%~v@t~a" indent r)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Utilities to manipulate data-tables
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod column-of ((d data-table) (attribute t))
  (position attribute (column-names d)))

(defmethod distinct-values ((d data-table) (a t))
  "Finds the unique values of attribute a in data table d."
  (let* ((attr-col (column-of d a)))
    (delete-duplicates
     (mapcar #'(lambda (row) (elt row attr-col)) (rows d)))))

(defmethod rows-matching ((d data-table) (a t) (v t))
  "Returns a list of the rows of data-table d for which the attribute a has
value v."
  (let ((the-col-num (column-of d a)))
    (remove-if #'(lambda (row) (not (eql v (elt row the-col-num))))
               (rows d))))

(defmethod split-table-by-attribute ((d data-table) (a t))
  "Yields an alist, each entry of which associates one of the values of the
attribute a with that subset of rows of d which match the particular value
for attribute a.  The total size of all the tables equals the size of the
original table."
  (mapcar #'(lambda (distinct-val)
              (cons distinct-val
                    (rows-matching d a distinct-val)))
          (distinct-values d a)))

(defmethod total-n ((d data-table))
  (length (rows d)))

(defmethod make-subset-data-table ((d data-table) subset-rows)
  (make-instance 'data-table
    :column-names (column-names d)
    :name-column (name-column d)
    :result-column (result-column d)
    :rows subset-rows))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Calculation of disorder in data-tables
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun -plogp (p) (- (* p (log p 2))))

(defmethod disorder ((d data-table))
  "Calculates the disorder (defined in the notes, page 357) of data table d."
  (if (not (= (cached-disorder d) -1.0))
    (cached-disorder d)
    (let ((n (total-n d))
          (subsets (split-table-by-attribute d (result-column d))))
      (let ((ans
             (reduce #'+ (mapcar #'(lambda (alist-entry)
				     (-plogp (/ (length (cdr alist-entry)) n)))
				 subsets))))
        (setf (cached-disorder d) ans)
        ans))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Calculation of disorder in id-trees
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod total-n ((tree id-tree))
  "Yields the total number of samples in tree."
  (let ((n 0))
    (dolist (b (branches tree))
      (incf n (total-n (cdr b))))
    n))

(defmethod disorder ((tree id-tree))
  "Calculates the disorder of an id-tree non-terminal node."
  (let ((dis 0.0) 
        (n (total-n tree)))
    (dolist (b (branches tree))
      (incf dis (* (/ (total-n (cdr b)) n)
                   (disorder (cdr b)))))
    dis))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Classification runs a sample through an id-tree and determines its result,
;;; which is simply the collection of cases that it matches.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod classify ((tree id-tree) (d data-table) (item t))
  "For item being a row with same column labels as data-table d,
returns the data-table leaf of the id tree into which this sample
item classifies."
  ;; For non-terminals, we must check the appropriate attribute of the item and
  ;; use it to choose the subtree.
  (let ((k (column-of d (attribute tree))))
    (when (null k) 
      (error "Attribute ~a from id-tree not in data table ~s" (attribute tree) d))
    (let ((a (assoc (elt item k) (branches tree))))
      (when (null a)
        (error "Value ~a for attribute ~a missing from id-tree"
               (elt item k) (attribute tree)))
      (classify (cdr a) d item))))

(defmethod classify ((tree data-table) (d data-table) (item t))
  ;; For the case where we have reached a terminal node of the id-tree, simply
  ;; return it as the classification.  This makes sense because the elements
  ;; in it will be homogeneous.
  (declare (ignore d item))
  tree)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Constructing the id-tree
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod make-id-tree-node ((d data-table) (a t) attrs-remaining)
  "Makes a new id-tree node which splits the data-table d on attribute a,
unless it is already homogeneous."
  (let ((split (split-table-by-attribute d a)))
    (if (null (cdr split))              ; homogeneous d
      d
      (make-instance 'id-tree
        :attribute a
        :remaining (remove a (or attrs-remaining (column-names d)))
        :branches (mapcar #'(lambda (assn)
                              (cons (car assn)
                                    (make-subset-data-table d (cdr assn))))
                          split)))))

(defun nearly-zerop (x)
  (< (abs x) 0.000001))

(defmethod make-id-tree-recursive ((d data-table) attrs-remaining)
  "Create an id-tree from the data-table d, using the best of the remaining
possible attributes (or all)"
  (if (nearly-zerop (disorder d))       ; Good enough; nothing left to do
    d
    (let* (;; Figure out all the remaining attributes on which to consider
           ;; splitting d
           (attrs (or attrs-remaining (column-names d)))
	   (attrs2 (cond ((equal attrs (list (result-column d)))
			  attrs)
			 (t
			  (remove (result-column d) attrs))))
           
           ;; Now construct each of the possible splits, but only one level down.
           ;; Note that we could guarantee an optimal id-tree if we actually
           ;; generated all the possible full id-trees here instead of just
           ;; splitting once, but of course that would be computationally terrible.
           (possible-id-trees
            (mapcar #'(lambda (attr)
                        (make-id-tree-node d attr attrs))
                    attrs2))
           
           ;; Compute the disorder associated with each split
           (disorders (mapcar #'disorder possible-id-trees)))
      
      ;; Now we select the locally-best (by one-step lookahead) split and then
      ;; further split it, if necessary.
      (if (null (cdr possible-id-trees))   ; there's only one, so it's the best!
        (car possible-id-trees)
        (do ((best-disorder (car disorders))
             (best-split (car possible-id-trees))
             (splits-remaining (cdr possible-id-trees) (cdr splits-remaining))
             (disorders-list (cdr disorders) (cdr disorders-list)))
            ((null splits-remaining)
             (split-recursively best-split))
          (when (< (car disorders-list) best-disorder)
            (setq best-split (car splits-remaining)
                  best-disorder (car disorders-list))))))))

(defmethod make-id-tree ((d data-table))
  "Create an id-tree from the data-table d, using the best of the remaining
possible attributes (or all)"
  (make-id-tree-recursive d (remove (name-column d) (column-names d))))

(defmethod split-recursively ((tree id-tree))
  "Given an id-tree node all of whose branches are in fact data tables,
as produced by make-id-tree, recursively consider splitting each of them into
an id-tree."
  (let ((attrs-remaining (remaining tree)))
    (make-instance 'id-tree
		   :attribute (attribute tree)
		   :remaining attrs-remaining
		   :branches (mapcar #'(lambda (assn)
					 (cons (car assn)
					       (make-id-tree-recursive 
						(cdr assn) attrs-remaining)))
				     (branches tree)))))

(defmethod split-recursively ((d data-table))
  "If split-recursively is called on a data-table, that means that at the
next-higher level, make-id-tree-node found it to be homogeneous on the attribute
of interest there, and it was still selected as the best attribute to split on.
In this case, a further split does not make sense."
  d)

(provide 'id3)
(provide "id3")
